/**
 * Routines for parsing and displaying DCC packets received from the DCC Monitor
 * board.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdint.h>
#include "global.h"
#include "display.h"
#include "dcc.h"

/**
 * Dump the packet contents if DCC verbosity is set
 */
static void dump_hex_data_if_verbose (struct s_frame *frame) {
  if (verbosity & VERB_DCC) {
    dump_hex_data (frame);
  }
}

/**
 * Parse a DCC command conforming to the "DCC Multi Function Digital Decoders"
 * format.
 * Source: Chapter C from RP-9.2.1 DCC Extended Packet Formats
 * Instruction Packets for Multi Function Digital Decoders
 *
 * Argument start: points to the first byte of the command, immediately
 * following the address contained in the DCC frame. start is  an array index
 * in struct s_frame.data.
 *
 * returns: number of bytes parsed from start, or -1 when an error occured and
 * hex_dumping has already been done. If no error occured, no trailing newline
 * has been printed (up to the caller).
 */
static int parse_DCC_multi_function_instruction (struct s_frame *frame, size_t start) {
  if (frame->len - start == 1) {
    // That's just the parity byte
    printf (" Unknown data (missing)\n");
    dump_hex_data_if_verbose (frame);
    return -1;
  }
  if ((frame->data[start] & 0xc0) == 0x40) {
    /* Command form: 01DCSSSS (binary)
     * Source: S-9.2 DCC Communications Standard
     * Locomotive speed, D = direction
     */
    uint8_t speed;

    printf (" Locomotive");
    // C is the least significant bit (though it might be the headlight, but
    // that's up to the user to recognise)
    speed = ((frame->data[start] & 0xf) << 1) | ((frame->data[start] & 0x10) >> 4);
    switch (speed) {
      case 0:
        printf (" stop");
        break;
      case 1:
        printf (" stop (I)"); // Means: locomotive may ignore direction
        break;
      case 2:
        printf (" emergency stop");
      case 3:
        printf (" emergency stop (I)");
      default:
        printf (" speed %u", speed - 3);
    }
    // Direction
    if (frame->data[1] & (1 << 5)) {
      printf (" forward");
    } else {
      printf (" reverse");
    }
    return 1;
  } else {
    printf (" Unknown command\n");
    dump_hex_data (frame);
    return -1; 
  }
}

/**
 * Parse a DCC command conforming to the "DCC Accessory Decoders"
 * format.
 * Source: Chapter D from RP-9.2.1 DCC Extended Packet Formats
 * Accessory Decoder Packet Formats
 */
static void parse_dcc_accessory_decoder_instruction (struct s_frame* frame) {
  uint16_t address;

  if (frame->len < 3) {
    printf (" Invalid Accessory Decoder command!\n");
    dump_hex_data (frame);
  }

  if (frame->data[1] & 0x80) {
    // Basic Accessory Decoder Packet

    /* Chapter D of RP 9.2.1 mentions a different addressing system from what I
     * see when I control turnouts from the Lenz LH100 handset. The real
     * difference is in the first byte; the second byte is principally the same
     * although a turnout has 1 output pair and a general Accessory Decoder
     * address has 4 pairs, mapping 4 turnouts on one Accessory Decoder address.
     * Also, turnouts occupy only half of the address space.
     *
     * I implemented the numbering from the handset for the address space
     * occupied by turnouts 1-1024 (I hope).
     *
     * The format seems to be:
     *
     * 10AAAAAA 11AACAAD
     *
     * Where:
     *
     * The first byte: the A's are bits 2-7 of the turnout address, plus 1! So
     * if bits 2-7 of the turnout address are all 0, the byte is 10000001. If
     * bits 2-7 of the address are all 1, the byte is 10000000.
     *
     * The second byte: bits 1-2 are bits 0-1 of the turnout address (least
     * significant).
     * Bits 4-5 are bits 8-9 of the turnout address (most significant), but in
     * one's complement.
     *
     * Also, what I refer to above as the turnout address, is 0-based, whereas
     * the numbering used on the handset in 1-based, so add 1.
     *
     * C: 1: activate, 0:deactivate. Turnouts are always activate, but it's
     * shown nonetheless.
     * D: 1: key "+", 0: key "-".
     *
     * When the second byte has the form 10AACDDD (notice the zero at bit 6, an
     * inverted address bit), we arrive in that part of the address space that
     * would lead to turnouts numbered 1025 and higher. I parse that as
     * specified in RP-9.2.1 chapter D.
     */

    if (frame->data[1] & 0x40) {
      // Turnout address space

      // Get the bits from the two bytes
      address = (((frame->data[1] & 0x30) ^ 0x30) << 4)
        | (((frame->data[0] - 1) & 0x3f) << 2)
        | ((frame->data[1] & 0x06) >> 1);
      
      // Zero-based to one-based
      address++;

      printf (" Accessory Decoder, turnout %u:", address);

      if (frame->len > 3) {
        if ((frame->data[2] & 0xf0) == 0xe0) {
          // Configuration Variable Access Instruction
          printf (" CV Access Instruction, not interpreted\n");

        } else {
          printf (" Invalid command!\n");
        }
        
        dump_hex_data (frame);
        return;
      }

      if (frame->data[1] & 0x01) {
        printf (" +");
      } else {
        printf (" -");
      }

      // And C is (de)activation
      if (frame->data[1] & 0x08) {
        printf (" activated\n");
      } else {
        printf (" deactivated\n");
      }

    } else {
      // Not a turnout address (it's a higher address), parse according to
      // RP-9.2.1.

      if ((frame->data[0] == 0xbf) && (frame->data[1] & 0xf0) == 0x80) {
        // Address form: 10111111 1000CDDD
        printf (" Accessory Decoder Broadcast:");

      } else {
        // Address form: 10AAAAAA 1AAACDDD
        // A's in second byte are most significant and in one's complement
        address = (frame->data[0] & 0x3f) | (((frame->data[1] & 0x70) ^ 0x70) << 2);
        printf (" Accessory Decoder %u:", address);
      }

      /* Of second byte: 1AAACDDD, the most significant two D's indicate which
       * pair of the output pairs is switched. The least significant D indicates
       * which output in a pair.
       */
      printf (" Pair %u output %u", (frame->data[1] & 0x06) >> 1, frame->data[1] & 0x01);

      if (frame->len > 3) {
        if ((frame->data[2] & 0xf0) == 0xe0) {
          // Configuration Variable Access Instruction
          printf (": CV Access Instruction, not interpreted\n");

        } else {
          printf (": Invalid command!\n");
        }
        
        dump_hex_data (frame);
        return;
      }

      // And C is (de)activation
      if (frame->data[1] & 0x08) {
        printf (" activated\n");
      } else {
        printf (" deactivated\n");
      }
    }
    
    dump_hex_data_if_verbose (frame);
    return;

  } else if ((frame->data[1] & 0x89) == 0x81) {
    // Extended Accessory Decoder Control Packet

    if (frame->len < 4) {
      printf (" Invalid Extended Accessory Decoder command!\n");
      dump_hex_data (frame);
      return;
    }

    if ((frame->data[0] == 0xbf) && frame->data[1] == 0x07) {
      // Address form: 10111111 00000111
      printf (" Extended Accessory Decoder Broadcast:");

    } else {
     // Address form: 10AAAAAA 0AAA0AA1
     // I assume the A's in the second byte are also in one's complement, seeing
     // the broadcast address above. RP-9.2.1 does not mention it!

     address = ((frame->data[0] & 0x3f) << 5)
       | ((((frame->data[1] & 0x70) >> 2) | ((frame->data[1] & 0x06) >> 1)) ^ 0x1f);
     printf (" Extended Accessory Decoder %u:", address);
    }

    if (frame->len > 4) {
        if ((frame->data[2] & 0xf0) == 0xe0) {
          // Configuration Variable Access Instruction
          printf (" CV Access Instruction, not interpreted\n");

        } else {
          printf (" Invalid command!\n");
        }
        
      dump_hex_data (frame);
      return;
    }

    if ((frame->data[2] & 0xe0) != 0x00) {
      // Third byte not of form 000XXXXX
      printf (" Invalid command!\n");
      dump_hex_data (frame);
      return;
    }

    if (frame->data[2] == 0x00) {
      printf (" absolute stop aspect\n");
    } else {
      // Third byte of form: 000XXXXX where the X..X is the aspect control
      printf (" aspect control %u\n", frame->data[2] & 0x1f);
    }

    dump_hex_data_if_verbose (frame);
    return;

  } else if ((frame->data[1] & 0x8c) == 0x0c) {
   /* Address form: 10AAAAAA 0AAA11VV
    * Deprecated Accessory Decoder Configuration Variable Access
    * Instruction.
    * Source: Appendix A from RP-9.2.1 DCC Extended Packet
    * Formats.
    */
   printf (" Deprecated Accessory Decoder CV Access Instruction, not interpreted\n");
   dump_hex_data (frame);
   return;
  }

  printf (" Invalid Accessory Decoder command!\n");
  dump_hex_data (frame);
  return;
}

/**
 * Parses a DCC frame and displays it on stdout
 */
int parse_dcc (struct s_frame *frame) {
  uint8_t parity;
  int retval;
  
  start_output_line (frame);
  printf (", proto DCC:");
  if (frame->len == 0) {
    // Empty DCC packets don't exist. An error-free monitoring board cannot
    // produce it.
    printf (" Invalid packet length 0 (monitor or communication error)!\n");
    return 0;
  }

  if (frame->len == 1) {
    // Just a parity byte?
    printf (" Unknown command: packet is only 1 byte\n");
    dump_hex_data (frame);
    return 0;
  }
  
  // Check parity
  parity = frame->data[frame->len - 1];
  for (int i=0; i < frame->len - 1; i++) {
    parity ^= frame->data[i];
  }
  if (parity != 0) {
    printf (" DCC Parity error!\n");
    dump_hex_data_if_verbose (frame);
    return 0;
  }

  if (!(frame->data[0] & (1 << 7))) {
    // Basic addres command form: 0AAAAAAA ... (binary)
    printf (" 7-bit address %u:", frame->data[0]);
    if (frame->len == 2) {
      printf (" Unknown data (missing)\n");
      dump_hex_data_if_verbose (frame);
      return 0;
    }
    // Decode command for Multi Function Digital Decoders
    retval = parse_DCC_multi_function_instruction (frame, 1);
    if (retval == -1) {
      // Error, but dumping hex data has already been taken care of
      return 0;
    } else if (retval + 2 != frame->len) {
      printf (", trailing DCC bytes not interpreted\n");
      dump_hex_data (frame);
      return 0;
    } else {
      putchar ('\n');
    }

  } else if (frame->data[0] == 0) {
    // Broadcast address
    printf (" Broadcast:");
    // Decode command for Multi Function Digital Decoders
    retval = parse_DCC_multi_function_instruction (frame, 1);
    if (retval == -1) {
      // Error, but dumping hex data has already been taken care of
      return 0;
    } else if (retval + 2 != frame->len) {
      printf (", trailing DCC bytes not interpreted\n");
      dump_hex_data (frame);
      return 0;
    } else {
      putchar ('\n');
    }
  } else if (frame->data[0] == 0xff) {
    if (frame->len == 3 && frame->data[1] == 0) {
      // DCC packet: 11111111 00000000 (binary)
      // Source: S-9.2 DCC Communications Standard
      printf (" Digital Decoder Idle Packet\n");
      dump_hex_data_if_verbose (frame);
      return 0;
    }
    printf (" Unknown command\n");
    dump_hex_data (frame);

  } else if ((frame->data[0] & 0xc0) == 0xc0) {
    /* 14-bit extended address form: 11AAAAAA AAAAAAAA
     * Source: Chapter C from RP-9.2.1 DCC Extended Packet Formats
     * It is already established the first byte != 11111111
     */
    uint16_t address;
    
    if (frame->len < 3) {
      printf (" Invalid 14-bit DCC address!\n");
      dump_hex_data (frame);
      return 0;
    }

    address = ((frame->data[0] & 0x3f) << 8) | frame->data[1];
    printf (" 14-bit address %u:", address);

    // Decode command for Multi Function Digital Decoders
    retval = parse_DCC_multi_function_instruction (frame, 2);
    if (retval == -1) {
      // Error, but dumping hex data has already been taken care of
      return 0;
    } else if (retval + 3 != frame->len) {
      printf (", trailing DCC bytes not interpreted\n");
      dump_hex_data (frame);
      return 0;
    } else {
      putchar ('\n');
    }

  } else if ((frame->data[0] & 0xc0) == 0x80) {
    // Accessory Decoder address form: 10AAAAAA
    // Source: Chapter D from RP-9.2.1 DCC Extended Packet Formats

    parse_dcc_accessory_decoder_instruction (frame);
    return 0;
  } else {
    printf (" Unknown command\n");
    dump_hex_data (frame);
    return 0;
  }
  
  dump_hex_data_if_verbose (frame);
  return 0;
}

