/**
 * Default handlers:
 * Unknown protocols
 * Error messages
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include "global.h"
#include "display.h"
#include "comm_proto.h"
#include "default_handlers.h"

/**
 * The default protocol handler, which gets called for any protocol that is not
 * handled.
 *
 * Reports an error to the user. Returns with zero to end protocol handler chain
 * traversal.
 */
int unknown_proto_handler (struct s_frame* frame) {
  start_output_line (frame);
  printf (": Unknown protocol %u!\n", frame->proto);
  dump_hex_data (frame);
  return 0;
}

/**
 * The default error handler
 *
 * Displays an error message for the errors returned by comm_get_frame() and
 * returns 0 to end the error handler chain.
 *
 * Errors END_INPUT and E_GARBAGE end the program.
 */
int default_err_handler (int error, struct s_frame* frame) {
  switch (error) {
    case END_INPUT:
      printf ("End of input stream.\n");
      exit (0);

    case E_GARBAGE:
      printf ("Too much garbage, exiting!\n");
      exit (1);

    case E_BIGFRAME:
      start_output_line (frame);
      printf (": Frame too big!\n");
      if (verbosity & VERB_ERR) {
        printf ("\tproto: %d", frame->proto);
        dump_hex_data (frame);
      }
      break;

    case E_NULLFRAME:
      start_output_line (frame);

      if (frame->proto == 0) {
        // These are forwarded idle frames
        printf  (": Protocol error: forwarded idle frame!\n");
      
      } else {
        printf (": Protocol error: incomplete frame (protocol: %u)!\n", frame->proto);
      }
      break;

    case E_TINYFRAME:
      start_output_line (frame);
      printf (": Protocol error: frame too small!\n");
      if (verbosity & VERB_ERR) {
        printf ("\tproto: %d", frame->proto);
        dump_hex_data (frame);
      }
      break;

    case E_PARITY:
      start_output_line (frame);
      printf (": Frame parity error!\n");
      if (verbosity & VERB_ERR) {
        printf ("\tproto: %d", frame->proto);
        dump_hex_data (frame);
      }
      break;
  }

  return 0;
}
