/**
 * Parsing routines for the RS-bus protocol.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include "global.h"
#include "display.h"
#include "rs.h"


/**
 * RS-bus frames have the following format:
 *
 * Normal frames: two bytes, address and a byte of data
 *
 * Address is the address of the responder on the RS-bus. High bit always
 * cleared (address range 0-127, RS-bus uses 1-128)
 *
 * Data format (bits): P T1 T0 N D3 D2 D1 D0
 * P      : bit parity, odd
 * T1 T0  : responder type:
 *          0 0 : switching receiver, no responder (is this useful?)
 *          0 1 : switching receiver with responder
 *          1 0 : stand-alone responder
 *          1 1 : reserved
 * N      : Nibble bit, 0 = lower nibble, 1 = higher nibble
 * D3..D0 : input pin state, 0 = passive, 1 = active
 *          N=0 : D3..D0 = E4..E1
 *          N=1 : D3..D0 = E8..E5
 *
 * -------------------------------
 * 
 * Special frames all start with a byte with the high bit set (to differentiate
 * them from normal frames):
 *
 * One byte frame:
 * 0x83 (RS_ADDR_ERR)     Addressing error: we observed more than 130 address
 *                        pulses. This message is sent only once in the lifetime
 *                        of the monitor process.
 *
 * Two byte frames:
 * 0x80 (RS_ZERO_ADDR) data
 *      Databyte sent after the address pulse for address 0; this is not a
 *      regular address.
 *
 * 0x81 (RS_HIGH_ADDR) data
 *      Databyte sent "from" address 129 or higher; not a regular address.
 *
 * 0x82 (RS_FRAME_ERR) address
 *      The specified (0-based) address sent data with a framing error (low
 *      endbit). Data is not included, just the 0-based address. Two special
 *      addresses exist:
 *      0xFF  The non-regular address 0 (cf. RS_ZERO_ADDR)
 *      0x80  The non-regular address 129 or higher (cf. RS_HIGH_ADDR)
 */

/**
 * Error codes sent by monitoring board.
 */
#define RS_ZERO_ADDR 0x80
#define RS_HIGH_ADDR 0x81
#define RS_FRAME_ERR 0x82
#define RS_ADDR_ERR 0x83

/**
 * Definitions for the bitfields in the databyte of a normal RS packet
 */
#define RS_TBITS ((1 << 6) | (1 << 5)) // The T1/T0 bits
#define RS_NO_RESP_TYPE 0  // Switching receiver, no responder
#define RS_SWITCH_TYPE (1 << 5) // Switching receiver with responder
#define RS_STAND_ALONE_TYPE (1 << 6) // Stand-alone responder
#define RS_RESERVED_TYPE ((1 << 6) | (1 << 5)) // Reserved responder type
#define RS_NIBBLE (1 << 4) // The nibble bit

struct rs_state {
  /**
   * A responder reports independently the state of the higher and lower nibble.
   * This variable is the bitwise or of the two values:
   * 0xF0 : We've seen the state of the higher nibble input pins before
   * 0x0F : The same for the lower nibble
   */
  uint8_t seen;
  /**
   * Last state of the input pins of the responder, MSB = E8, LSB = E1, 
   * 1 = active, 0 = passive.
   */
  uint8_t state;
};

/**
 * Two-dimensional array of responder states.
 * statelist[MA][RA] :
 * MA = Monitor address (from struct s_frame.addr)
 * RA = Responder address (RS-bus address of responder connected to the RS-bus
 *      monitored by monitor number MA); 0-based
 */
static struct rs_state **statelist;

/**
 * Initialise RS-bus protocol parser.
 *
 * Initialises state table
 * 
 * Needs to be called once before parsing any RS-bus frames.
 */
void init_rs () {
  // Allocate an array for all possible monitors
  statelist = malloc (MAX_ADDRS * sizeof (struct rs_state*));

  if (statelist == NULL) {
    fprintf (stderr, "Out of memory!\n");
    exit (1);
  }

  // And fill it with NULLs
  for (int i = 0; i < MAX_ADDRS; i++) {
    statelist[i] = NULL;
  }
}

/**
 * Checks the parity for an RS-bus data byte
 *
 * Returns 0 when parity fails, non-zero when correct
 */
static int rs_check_parity (uint8_t data) {
  uint8_t parity;

  parity = 0;
  while (data != 0) {
    parity ^= data & 1;
    data >>= 1;
  }
  return parity;
}

/**
 * Prints the active and passive RS pins on stdout, and marks with stars those
 * pins that changed state compared to old_data.
 *
 * old_data has the old nibble state in it's lower nibble, other nibble is
 * don't-care.
 * new_data has the new nibble state in it's lower nibble, and the RS_NIBBLE bit
 * indicates which nibble it is. Rest of the bits is don't-care.
 */
static void print_rs_pin_changes (const uint8_t old_data, const uint8_t new_data) {
  /**
   * Difference y-x between Dx and Ey (1 or 5)
   * Dx = databit in responder data
   * Ey = input pin on responder
   */
  uint8_t pin_nr_offs; 
  uint8_t old_data_temp, new_data_temp;

  if (new_data & RS_NIBBLE) {
    // Higher nibble
    pin_nr_offs = 5;
  } else {
    // Lower nibble
    pin_nr_offs = 1;
  }

  old_data_temp = old_data;
  new_data_temp = new_data;

  // First report the active inputs
  if (new_data_temp & 0x0F) {
    // There are active pins

    printf (" Active:");
    for (uint8_t i = 0; i < 4; i++) {
      if (new_data_temp & 1) {
        // This pin is active

        if (old_data_temp & 1) {
          // It was already active
          printf (" E%u", i + pin_nr_offs); // Prints the input pin name
        } else {
          // It wasn't active before, mark it as changed with stars
          printf (" *E%u*", i + pin_nr_offs); // Prints the input pin name with markers
        }
      }

      // Next bit
      new_data_temp >>= 1;
      old_data_temp >>= 1;
    }

    // Get the original data back, we consumed it
    old_data_temp = old_data;
    new_data_temp = new_data;
  }

  // Now report the passive inputs
  if ((new_data_temp & 0x0F) ^ 0x0F) {
    // There are passive pins
    
    if (new_data_temp & 0x0F) {
      // There are active pins, so use comma and lowercase
      printf (", passive:");
    } else {
      // There are only passive pins, so we "start a sentence"
      printf (" Passive:");
    }

    for (uint8_t i = 0; i < 4; i++) {
      if (!(new_data_temp & 1)) {
        // This pin is passive

        if (!(old_data_temp & 1)) {
          // It was already passive
          printf (" E%u", i + pin_nr_offs); // Prints the input pin name
        } else {
          // It wasn't passive before, mark it as changed with stars
          printf (" *E%u*", i + pin_nr_offs); // Prints the input pin name with markers
        }
      }

      // Next bit
      new_data_temp >>= 1;
      old_data_temp >>= 1;
    }
  }
}
  
/**
 * Parses the RS-bus frame and displays it on stdout
 */
int parse_rs (struct s_frame* frame) {
  uint8_t address; // The address of the responder on the RS-bus
  uint8_t data; // The databyte sent by the responder
  uint8_t error; // The error code sent by the monitoring board
  uint8_t state_compare; // The last state of the input pins (one nibble)
  struct rs_state *my_statelist; // Responder state array for this monitoring board
  struct rs_state *state; // State of this responder

  start_output_line (frame);
  printf (", proto RS:");

  if (frame->len == 0) {
    // Such frames don't exist
    printf (" Invalid packet length 0 (monitor or communication error)!\n");
    return 0;
  }

  if (!(frame->data[0] & (1 << 7))) {
    // Regular data packet

    if (frame->len != 2) {
      // Impossible
      printf (" Invalid packet contents (monitor or communication error)!\n");
      dump_hex_data (frame);
      return 0;
    }

    address = frame->data[0];
    data = frame->data[1];

    // 0-based address to 1-based address
    address++;

    printf (" Address %u:", address);

    // Check parity
    if (!rs_check_parity (data)) {
      // Parity error
      printf (" Parity error!\n");
      if (verbosity & VERB_RS) {
        dump_hex_data (frame);
      }
      return 0;
    }

    // Look for available state info
    my_statelist = statelist[frame->addr];

    if (my_statelist == NULL) {
      // We never had any RS-packet from this monitor before, hence there is
      // no statelist. Build it.
      
      // 128 possible responders, one state for each
      my_statelist = malloc (128 * sizeof (struct rs_state));

      if (my_statelist == NULL) {
        fprintf (stderr, "Out of memory!\n");
        exit (1);
      }

      // Set them all to unseen
      for (uint8_t i = 0; i < 128; i++) {
        my_statelist[i].seen = 0;
      }

      // Put it in the global list
      statelist[frame->addr] = my_statelist;
    }

    // State for this responder
    state = &my_statelist[address - 1];

    if (data & RS_NIBBLE) {
      // This is the higher nibble

      if (!(state->seen & 0xF0)) {
        // This is the first time we see this responder nibble.
        // We report such a state as if nothing changed.
        state->state &= 0x0F; // Clear any rubble bits
        state->state |= (data & 0x0F) << 4; // Set "old state" from current state
        state->seen |= 0xF0; // We've seen it now
      }

      // Set state_compare to contain the old state in the lower nibble
      state_compare = state->state >> 4;

    } else {
      // This is the lower nibble

      if (!(state->seen & 0x0F)) {
        // This is the first time we see this responder nibble.
        // We report such a state as if nothing changed.
        state->state &= 0xF0; // Clear any rubble bits
        state->state |= data & 0x0F; // Set "old state" from current state
        state->seen |= 0x0F; // We've seen it now
      }

      // Set state_compare to contain the old state in the lower nibble
      state_compare = state->state & 0x0F;
    }

    // Print the pin states and changes to states
    print_rs_pin_changes (state_compare, data);

    // End of output line
    putchar ('\n');

    // Get the original data back, we consumed it
    data = frame->data[1];

    // Save the current state of the input pins
    if (data & RS_NIBBLE) {
      // This is the higher nibble
      state->state &= 0x0F; // Clear the old bits
      state->state |= (data & 0x0F) << 4; // Set state
    } else { 
      // This is the lower nibble
      state->state &= 0xF0; // Clear the old bits
      state->state |= data & 0x0F; // Set state
    }

    if (verbosity & VERB_RS) {
      dump_hex_data (frame);
    }

    return 0;
  }

  // Special frames (high bit of first byte set)
  error = frame->data[0];
  switch (error) {
    case RS_ZERO_ADDR:
      if (frame->len != 2) {
        // Impossible
        printf (" Invalid packet contents (monitor or communication error)!\n");
        dump_hex_data (frame);
        return 0;
      }
      
      data = frame->data[1];

      // Check parity
      if (!rs_check_parity (data)) {
        // Parity error
        printf (" Invalid address 0: Parity error!\n");

      } else {
        // Parity correct
        printf (" Invalid address 0:");
        
        // We don't keep state information on invalid addresses; just report state
        // without changes
        print_rs_pin_changes (data, data);
        printf (" !\n");
      }
      break;
      
    case RS_HIGH_ADDR:
      if (frame->len != 2) {
        // Impossible
        printf (" Invalid packet contents (monitor or communication error)!\n");
        dump_hex_data (frame);
        return 0;
      }
      
      data = frame->data[1];

      // Check parity
      if (!rs_check_parity (data)) {
        // Parity error
        printf (" Invalid address >=129: Parity error!\n");

      } else {
        // Parity correct
        printf (" Invalid address >=129:");
        
        // We don't keep state information on invalid addresses; just report state
        // without changes
        print_rs_pin_changes (data, data);
        printf (" !\n");
      }
      break;
      
    case RS_FRAME_ERR:
      if (frame->len != 2) {
        // Impossible
        printf (" Invalid packet contents (monitor or communication error)!\n");
        dump_hex_data (frame);
        return 0;
      }
      
      address = frame->data[1];
      if (address == 0xFF) {
        // Zero address
        printf (" Invalid address 0:");

      } else if (address == 0x80) {
        // High address
        printf (" Invalid address >=129:");

      } else {
        // Convert 0-based address to 1-based address
        address++;
        printf (" Address %u:", address);
      }

      printf (" Framing error!\n");
      break;

    case RS_ADDR_ERR:
      if (frame->len != 1) {
        // Impossible
        printf (" Invalid packet contents (monitor or communication error)!\n");
        dump_hex_data (frame);
        return 0;
      }
     
      printf (" Seen to many address pulses!\n\t(This error is reported only once)\n");
      break;

    default:
      // Unknown frame contents
      printf (" Invalid packet contents (monitor or communication error)!\n");
      dump_hex_data (frame);
      return 0;
  }

  if (verbosity & VERB_RS) {
    dump_hex_data (frame);
  }

  return 0;
}
