/**
 * These routines parse frames coming in from the DCC Monitor board and pass
 * them to the caller.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <limits.h>
#include "global.h"
#include "serial_comm_proto.h"

static FILE *input; // Input stream
static unsigned int total_byte_cnt = 0; // Total number of bytes received from input stream

/**
 * Set the given stream to be the input stream, which should be connected to the
 * DCC Monitor board.
 */
void sercomm_set_input_stream (FILE *stream) {
  input = stream;
}

/**
 * Return the total number of bytes read since startup
 */
unsigned int sercomm_get_byte_cnt () {
  return total_byte_cnt;
}

/**
 * Look for the start of a frame in the input stream
 * Verbosely reports how much bytes were skipped to get to the next start
 * @returns END_INPUT when the end of the input stream was reached
 *          E_GARBAGE if there was an incredible amount of non-start characters
 */
static int lock_stream () {
  unsigned int garbage_cnt; // Counts non-framestart bytes discarded
  int bytes_written; // How many characters to backspace over on the output
  int addr_and_proto; // Frame start byte
  
  garbage_cnt = 0;
  
  printf ("Locking on to stream... 0");
  bytes_written = 1;

  addr_and_proto = getc(input);
  if (addr_and_proto == EOF) {
    // End of input stream
    putchar ('\n');
    return END_INPUT;
  }
  total_byte_cnt++;

  while (!(addr_and_proto & (1 << 7))) {
    // Not the start of a frame

    // Limit checking
    if (garbage_cnt == UINT_MAX) {
      return E_GARBAGE;
    }

    garbage_cnt++;
    for (; bytes_written > 0; bytes_written--) {
      putchar ('\b'); // Backspace over old printed value
    }

    /* Note: because EOF < 0 this next line results in
     * an error of printf being ignored by the backspacing
     * for-loop
     */
    bytes_written = printf ("%u", garbage_cnt);
    addr_and_proto = getc (input);
    if (addr_and_proto == EOF) {
      // End of input stream
      putchar ('\n');
      return END_INPUT;
    }
    total_byte_cnt++;
  }
  putchar('\n');
  return addr_and_proto;
}

/**
 * Return the next frame from the input stream
 *
 * Return value is 0 for success, or:
 * END_INPUT: End of input stream reached
 * E_BIGFRAME: Frame received was too large
 * E_NULLFRAME: Just a frame start byte
 * E_TINYFRAME: Frame received was too small
 * E_PARITY: Frame failed parity check (frame parity != DCC parity!)
 */
int8_t sercomm_get_frame (struct s_frame *frame) {
  static int data; // The last read byte is passed on to next invocation
  uint8_t byte_count; // Count towards reception of a byte containing hi_bits
  uint8_t hi_bits = 0; // Keeps hi_bits until we know we can process it
                       // (initialised to prevent compiler warning)
  size_t bufp; // Pointer in frame databuffer
  uint8_t parity; // Parity checking for frame

  if (!(data & (1 << 7))) {
    // Not the start of a frame, first lock on to the stream
    data = lock_stream();
    if (data < 0) {
      // Pass on error
      return (int8_t) data;
    }
  }

  // Read in databytes
  bufp = 0;
  byte_count = 0;
  parity = (uint8_t) data; // Initialise parity (data contains the frame start byte right now)
  frame->proto = parity & 15;
  frame->addr = ((parity & 112) >> 4);

  data = getc (input);
  if (data == EOF) {
    // End of input stream
    return END_INPUT;
  }
  total_byte_cnt++;

  // Special case: if we received addr 0, proto 0 and the next is the start of a
  // frame, it's an Idle Frame, so skip over it
  while ((frame->addr == 0) && (frame->proto == 0) && (data & (1 << 7))) {
    parity = (uint8_t) data;
    frame->proto = parity & 15;
    frame->addr = ((parity & 112) >> 4);
    data = getc (input);
    if (data == EOF) {
      return END_INPUT;
    }
    total_byte_cnt++;
  }

  // Set reception time
  frame->time = time (NULL);

  while (!(data & (1 << 7))) {
    // Still in the frame
    byte_count++;
    parity ^= data;
    if (byte_count == 8) {
      /* This is probably the hi_bits byte
       * There is one exception: this could also be the parity byte. We know
       * which is which when we receive the next byte (if it's the start of a
       * frame, it was the parity byte)
       */
      hi_bits = data;
      data = getc (input);
      if (data == EOF) {
        return END_INPUT;
      }
      total_byte_cnt++;
      continue; // Next byte
      
    } else if (byte_count == 9) {
      // The previous byte was hi_bits
      for (size_t p = bufp - 7; p < bufp; p++) {
        // Set the high bit in the buffer correctly
        if ((hi_bits & 1)) {
          frame->data[p] |= (1 << 7);
        }
        // Shift hi_bits
        hi_bits >>= 1;
      }
      byte_count = 1;
    }

    if (bufp >= MAX_FRAME_SIZE + 2) {
      // Frame too large
      // Set length to max
      frame->len = MAX_FRAME_SIZE;
      return E_BIGFRAME;
    }

    // Place byte in buffer
    frame->data[bufp] = (uint8_t) data;
    bufp++;
    data = getc (input);
    if (data == EOF) {
      return END_INPUT;
    }
    total_byte_cnt++;
  }
  
  // A full frame has been received, but some processing still has to be done
  frame->len = bufp;

  /* Minimum length for a frame with data is 3 bytes (data, hi_bits, parity)
   * Frame without data is 1 byte (parity)
   * 0 bytes is invalid (only the special case of the Idle Frame, which is
   * already handled above).
   */
  if (frame->len == 0) {
    return E_NULLFRAME;
  } else if (frame->len == 2) {
    return E_TINYFRAME;
  }

  // Check parity (high bit ignored)
  if (parity & 127) {
    return E_PARITY; // Parity error
  }

  /* Substract parity from data
   * If byte_count == 8, parity was already dropped as it was seen as a
   * potential hi_bits byte
   */
  if (byte_count != 8) {
    frame->len--;
  }

  // Do we need to parse hi_bits?
  if (byte_count != 1) { // Parity byte was also counted
    // Yes
    uint8_t hi_bits;

    hi_bits = frame->data[frame->len - 1];
    for (size_t p = frame->len - ((size_t)byte_count) + 1; p < frame->len - 1; p++) {
      // Set the high bit in the buffer correctly
      if (hi_bits & 1) {
        frame->data[p] |= (1 << 7);
      }
      // Shift hi_bits
      hi_bits >>= 1;
    }
    // Substract hi_bits from frame
    frame->len--;
  }
  
  if (frame->len > MAX_FRAME_SIZE) {
    /* There is one possibility for this to happen: if MAX_FRAME_SIZE is a
     * multiple of 7, and the frame received is MAX_FRAME_SIZE + 1 bytes long,
     * the earlier test for maximum size will not have caught it because it
     * accounted for hi_bits and parity, whereas only parity was received
     */

    // Set length to max
    frame->len = MAX_FRAME_SIZE;
    return E_BIGFRAME;
  }
  return 0; 
}

