/**
 * Functions for integrating the "Communication forward" test from the firmware
 * into the PC program
 *
 * Frames received from the test running in the firmware are checked for
 * validity.
 
 * The communication forward test is inherently stateful; frames which are
 * missed (e.g., dropped) or deliberate errors which are not recognised
 * correctly cause the receiver to go out-of-sync with the board running the
 * test. This usually results in an impressive list of error messages. It is up
 * to the user (developer) to trace the cause of the error.
 *
 * A correctly run test however results in just a confirmation that everything
 * went okay. This is the strength of the automated test: verifying correct
 * handling of the test cases is very little work.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stddef.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "../global.h"
#include "../comm_proto.h"
#include "../handler_functions.h"
#include "../display.h"
#include "test_emul.h"
#include "test_comm_forward_handler.h"

/**
 * test_data_item.status keeps the status on the running test. Possible values
 * are:
 * 0 - Test okay so far
 * 1 - Test failed at some point
 * 2 - Test should have ended, but is still running
 */

/**
 * Management frame handler for management frames coming from the board to which
 * the board running the test is connected.
 *
 * Only frames which we expect are handled here, anything unexpected is passed
 * through to the next protocol handler.
 *
 * @returns 0 when we handled the frame because we expected it, non-zero
 * otherwise.
 *
 * Precondition: the test has been correctly initialized.
 */
static int parse_forward_test_manag (struct s_frame* frame) {
  struct test_data_item *current_test;
  uint8_t *forward_test_num; // Same as FORWARD_TEST_NUM in firmware

  // Get data for this test (test is running on board connected to the one
  // giving this management frame!)
  current_test = test_emul_get_data (frame->addr + 1);

  // Calling this function for a non-running test is a programming error
  assert (current_test != NULL);
  
  if (current_test->status == 2) {
    // Our emulation already signaled completion, don't handle these frames 
    return -1;
  }

  // Set forward_test_num correctly
  forward_test_num = (uint8_t*) ((void *)current_test->test_data);

  switch (*forward_test_num) {
    case 0:
      // Test if next board correctly generates "Chain Too Long" by sending it a
      // frame that appears to come from address 7.
      if (frame->len == 1 && frame->data[0] == 3) {
        // This is a "Chain Too Long" error, which we expected
        // Test succeeded
        if (verbosity & VERB_TEST) {
          start_output_line (frame);
          printf (": \"Communication forward test\" frame correctly triggered \"Chain too long\" error.\n");
          dump_hex_data (frame);
        }

      } else {
        // Not what we were expecting, don't handle it
        return -1;
      }
      break;

    case 1:
      // Test that next board correctly detects we overflow the input buffer
      // with one single frame. Obviously this depends on the input buffers
      // being equally big in this and the next board.
      
      if (frame->len == 1 && frame->data[0] == 2) {
        // This is a "Malformed packet" error, which we expected
        // Test succeeded
        if (verbosity & VERB_TEST) {
          start_output_line (frame);
          printf (": \"Communication forward test\" frame correctly triggered \"Malformed packet\" error.\n");
          dump_hex_data (frame);
        }
      } else {
        // Not what we were expecting, don't handle it
        return -1;
      }
      break;

    default:
      // We're not expecting a management frame from this board now, don't
      // handle
      return -1;
  }

  // Set next test
  (*forward_test_num)++;
  
  // We handled the frame
  return 0;
}

/**
 * Initialise the forward test running on the board with address addr,
 * using protocol proto.
 *
 * To be called when the "start comm. test" frame is received.
 *
 * Initialises states and attaches handlers to all types of frames and errors we
 * expect to see.
 *
 * Precondition: addr > 0 (running the forward test on board 0 does not work)
 */
void init_forward_test (const uint8_t addr, const uint8_t proto) {
  struct test_data_item *current_test;
  uint8_t *forward_test_num; // Same as FORWARD_TEST_NUM in firmware
 
  // Get the test environment
  current_test = test_emul_get_data (addr);

  // Calling this function for a non-running test is an error
  assert (current_test != NULL);

  forward_test_num = (uint8_t *) ((void *) current_test->test_data);

  // Initialise forward_test_num
  *forward_test_num = 0;

  // Register the handler that parses management protocol frames from
  // the board to which this board is connected
  reg_handler (addr - 1, 0, parse_forward_test_manag);
}

/** 
 * End the forward test and inspect result.
 *
 * Unregisters the handler registered by init_forward_test() 
 *
 * @returns 0 when everything is okay, and -2 when we were still expecting more
 * frames.
 *
 * We don't know if frames errored; no idea what the result would look like so
 * we can't check for it.
 * 
 * Precondition: the test is correctly initialized.
 */
int end_forward_test (const uint8_t addr) {
  struct test_data_item *current_test;
  uint8_t *forward_test_num; // Same as FORWARD_TEST_NUM in firmware
  int retval;

  // Get data for this test
  current_test = test_emul_get_data (addr);

  // Calling this function for a non-running test is an error
  assert (current_test != NULL);

  forward_test_num = (uint8_t *) ((void *) current_test->test_data);

  if (*forward_test_num != 2) {
    // We were still expecting frames
    retval = -2;
  } else {
    // Seems okay
    retval = 0;
  }

  /* The test is identified only by the address (always at most 1 test
   * running), so we can delete it without knowing the proto number.
   *
   * Unregister the handler that parses management protocol frames from
   * the board to which this board is connected.
   */
  unreg_handler (addr - 1, 0, parse_forward_test_manag);

  return retval;
}
