/**
 * Functions for integrating the "Communication protocol" test from the firmware
 * into the PC program
 *
 * Frames received from the test running in the firmware are checked for
 * validity; normal frames are emulated and compared, special frames (deliberate
 * errors) are checked if they were processed correctly by the PC program or the
 * monitoring board to which the board running the test is daisy-chained.
 *
 * The communication protocol test is inherently stateful; frames which are
 * missed (e.g., dropped) or deliberate errors which are not recognised
 * correctly cause the receiver to go out-of-sync with the board running the
 * test. This usually results in an impressive list of error messages. It is up
 * to the user (developer) to trace the cause of the error.
 *
 * A correctly run test however results in just a confirmation that everything
 * went okay. This is the strength of the automated test: verifying correct
 * handling of the test cases is very little work.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stddef.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "../global.h"
#include "../comm_proto.h"
#include "../handler_functions.h"
#include "../display.h"
#include "test_emul.h"
#include "comm_emul.h"
#include "test_comm_proto.h"
#include "test_comm_proto_glue.h"

/**
 * The same struct as used in test_comm_proto, so we can read it too.
 */
struct comm_test_var {
  uint8_t len;
  uint8_t bitpos;
};

/**
 * test_data_item.status keeps the status on the running test. Possible values
 * are:
 * 0 - Test okay so far
 * 1 - Test failed at some point
 * 2 - Test should have ended, but is still running
 */

/**
 * Frame handler for the test frames that are received.
 *
 * Correct test result of this frame is checked, and an error is printed if the
 * result is incorrect.
 * 
 * It should be bound to the protocol number that the board reported in the
 * "start comm. test" frame.
 *
 * Precondition: the test has been correctly initialized.
 */
static int parse_comm_test (struct s_frame* frame) {
  struct s_frame emul_frame;
  struct test_data_item *current_test;
  struct comm_test_var test_var;

  // Get data for this test
  current_test = test_emul_get_data (frame->addr);

  // Calling this function for a non-running test is a programming error
  assert (current_test != NULL);
  
  if (current_test->status == 2) {
    // Our emulation already signaled completion, but we're still getting test
    // frames. 
    start_output_line (frame);
    printf (": \"Communication test\" should have ended, but is still running!\n");
    dump_hex_data (frame);
    return 0;
  }

  // Set the context for the test
  test_emul_set_current_test (frame->addr);
  
  // Get the test variables as they are now, just before the current test
  memcpy (&test_var, &global_test_var, sizeof (struct comm_test_var));

  // Emulate the test in this frame
  set_comm_emul_buf (&emul_frame);

  // Avoid emulating one specific test which would trigger an assertion
  if (!(test_var.len == MAX_FRAME_SIZE + 1 && test_var.bitpos == 6)) {
    // Emulate the test
    if (do_comm_test (0)) {
      // This run sent the management frame, but here we are handling a test
      // frame. Something must have gone wrong.

      start_output_line (frame);
      printf (": \"Communication test\" should have ended, but is still running!\n");
      dump_hex_data (frame);
      current_test->status = 2;
      return 0;
    }
  }

  // Save the context for the test
  test_emul_save_current_test (frame->addr);

  if (test_var.len == MAX_FRAME_SIZE + 1) {
    // There are some special frames that need to be handled specially in this
    // part of the test

    // bitpos out of range means do_comm_test screwed up in a previous run and 
    // is a programming error
    assert (test_var.bitpos <= 6);

    start_output_line (frame);
    switch (test_var.bitpos) {
      case 1:
        // Deliberate parity error
        // Excludes the data from the parity to test that it is included in
        // calculation
        printf (": \"Communication test\" error: deliberate parity error in data part was not detected!\n\tGot:\n");
        dump_hex_data (frame);
        break;

      case 2:
        // Deliberate parity error
        // Excludes the hi-bits portion from the parity to test that it is
        // included in calculation
        printf (": \"Communication test\" error: deliberate parity error in hi-bits part was not detected!\n\tGot:\n");
        dump_hex_data (frame);
        break;

      case 3:
        // Deliberate parity error
        // Excludes the frame start byte from the parity to test that it is
        // included in calculation
        printf (": \"Communication test\" error: deliberate parity error in start byte part was not detected!\n\tGot:\n");
        dump_hex_data (frame);
        break;

      case 4:
        // Incomplete frame: just a frame start byte
        printf (": \"Communication test\" error: deliberate null frame error was not detected!\n\tGot:\n");
        dump_hex_data (frame);
        break;

      case 5:
        // Malformed packet: a 3-byte frame doesn't exist
        printf (": \"Communication test\" error: deliberate tiny frame error was not detected!\n\tGot:\n");
        dump_hex_data (frame);
        break;

      case 6:
        // Impossibly big frame
        
        printf (": \"Communication test\" error: deliberately too big frame was not detected!\n\tGot:\n");
        dump_hex_data (frame);
        // We need to manually change the test state because this test couldn't
        // be emulated
        test_var.bitpos = 7;
        memcpy (&(current_test->test_data), &test_var, sizeof (struct comm_test_var));
    }
    // Test failed
    current_test->status = 1;
    return 0;
  }
  
  // Now compare the received frame and the result of the emulation
  if (test_emul_compare_frames (frame, &emul_frame)) {
      // Frames the same
      if (verbosity & VERB_TEST) {
        start_output_line (frame);
        printf (": \"Communication test\" frame is okay.\n");
        dump_hex_data (frame);
      }
      return 0;
  }
  
  // Frames not the same
  start_output_line (frame);
  printf (": \"Communication test\" frame is not as expected!\n\tExpected:\n");
  dump_hex_data (&emul_frame);
  printf ("\tGot:\n");
  dump_hex_data (frame);

  // Set test status to failed
  current_test->status = 1;
  return 0;
}
/**
 * Management frame handler for management frames coming from the board to which
 * the board running the test is connected.
 *
 * Some frames sent by the communication test should cause the board to which it
 * is connected to send a management frame with an error code. These frames are
 * caught by this function. The emulation is updated to be in sync with the
 * board.
 *
 * Only frames which we expect are handled here, anything unexpected is passed
 * through to the next protocol handler.
 *
 * @returns 0 when we handled the frame because we expected it, non-zero
 * otherwise.
 *
 * Precondition: the test has been correctly initialized.
 */
static int parse_comm_test_manag (struct s_frame* frame) {
  struct s_frame emul_frame;
  struct test_data_item *current_test;
  struct comm_test_var *test_var;
  uint8_t test_retval;

  // Get data for this test (test is running on board connected to the one
  // giving this management frame!)
  current_test = test_emul_get_data (frame->addr + 1);

  // Calling this function for a non-running test is a programming error
  assert (current_test != NULL);
  
  if (current_test->status == 2) {
    // Our emulation already signaled completion, don't handle these frames 
    return -1;
  }

  // Set the context for the test
  test_emul_set_current_test (frame->addr + 1);
  
  // Emulate the test in this frame
  set_comm_emul_buf (&emul_frame);

  // Get test variables
  test_var = (struct comm_test_var*) ((void*) current_test->test_data);

  if (test_var->len == MAX_FRAME_SIZE + 1) {
    // This might be a packet we wish to catch

    switch (test_var->bitpos) {
      case 4:
        // Incomplete frame: just a frame start byte

        if (frame->len == 1 && frame->data[0] == 2) {
          // This is a "Malformed packet" error, which we expected
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame on daisy-chain correctly triggered \"Malformed packet\" error.\n");
            printf ("Test case: null frame.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not what we were expecting, don't handle it
          return -1;
        }
        break;

      case 6:
        // Impossibly big frame
        
        
        if (frame->len == 1 && frame->data[0] == 2) {
          // This is a "Malformed packet" error, which we expected
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame on daisy-chain correctly triggered \"Malformed packet\" error.\n");
            printf ("Test case: too big frame.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not the expected error, don't handle it
          return -1;
        }

        // We need to manually change the test state because this test can't
        // be emulated: it triggers an assertion
        test_var->bitpos = 7;
        return 0;

      default:
        // We're not expecting a management frame from this board now, don't
        // handle
        return -1;
    }

    // Emulate the test (discard results in emul_frame)
    test_retval = do_comm_test (0);

    // It would be a programming error if do_comm_test indicates the test
    // completed.
    assert (test_retval == 0);

    // Save the context for the test
    test_emul_save_current_test (frame->addr + 1);

    // We handled this frame
    return 0;
  }

  // We're not expecting a management frame from this board now, don't
  // handle

  return -1;
}

/**
 * Handle communication errors which are the result of running the communication
 * protocol test
 *
 * The communication protocol test deliberately generates errors; this handler
 * catches those deliberate errors and updates the emulation to be in sync with
 * the board.
 *
 * Only errors we expect are handled; others are passed on to the next error
 * handler (usually resulting in an error message to be printed)
 *
 * @returns 0 when the error was expected, non-zero otherwise
 *
 * Precondition: the test has been correctly initialized.
 */
static int parse_comm_test_err (int error, struct s_frame* frame) {
  struct s_frame emul_frame;
  struct test_data_item *current_test;
  struct comm_test_var *test_var;
  uint8_t test_retval;

  // Get data for this test
  current_test = test_emul_get_data (frame->addr);

  // Calling this function for a non-running test is a programming error
  assert (current_test != NULL);
  
  if (current_test->status == 2) {
    // Our emulation already signaled completion, we don't handle this error.
    return -1;
  }

  // Set the context for the test
  test_emul_set_current_test (frame->addr);
  
  // Emulate the test in this frame
  set_comm_emul_buf (&emul_frame);

  // Get test variables
  test_var = (struct comm_test_var*) ((void*) current_test->test_data);

  if (test_var->len == MAX_FRAME_SIZE + 1) {
    // This might be an error packet we wish to catch

    // bitpos out of range means do_comm_test screwed up in a previous run and 
    // is a programming error
    assert (test_var->bitpos <= 7);

    switch (test_var->bitpos) {
      case 1:
        // Deliberate parity error
        // Excludes the data from the parity to test that it is included in
        // calculation

        if (error == E_PARITY) {
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame correctly triggered parity error in data part.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not the expected error, don't handle it
          return -1;
        }
        break;

      case 2:
        // Deliberate parity error
        // Excludes the hi-bits portion from the parity to test that it is
        // included in calculation
        if (error == E_PARITY) {
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame correctly triggered parity error in hi-bits part.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not the expected error, don't handle it
          return -1;
        }
        break;

      case 3:
        // Deliberate parity error
        // Excludes the frame start byte from the parity to test that it is
        // included in calculation

        if (error == E_PARITY) {
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame correctly triggered parity error in start byte part.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not the expected error, don't handle it
          return -1;
        }
        break;

      case 4:
        // Incomplete frame: just a frame start byte
        // Test frames from addresses > 0 should be caught by the board to which
        // it is connected. Only trigger on tests running on board address 0.

        if (frame->addr == 0 && error == E_NULLFRAME) {
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame correctly triggered null frame error.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not the expected error, don't handle it
          return -1;
        }
        break;

      case 5:
        // Malformed packet: a 3-byte frame doesn't exist
        if (error == E_TINYFRAME) {
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame correctly triggered tiny frame error.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not the expected error, don't handle it
          return -1;
        }
        break;

      case 6:
        // Impossibly big frame
        
        if (error == E_BIGFRAME) {
          // Test succeeded
          if (verbosity & VERB_TEST) {
            start_output_line (frame);
            printf (": \"Communication test\" frame correctly triggered big frame error.\n");
            dump_hex_data (frame);
          }
        } else {
          // Not the expected error, don't handle it
          return -1;
        }

        // We need to manually change the test state because this test can't
        // be emulated: it triggers an assertion
        test_var->bitpos = 7;
        return 0;

      default:
        // This shouldn't trigger error; don't handle
        return -1;
    }

    // Emulate the test (discard results in emul_frame)
    test_retval = do_comm_test (0);

    // It would be a programming error if do_comm_test indicates the test
    // completed.
    assert (test_retval == 0);

    // Save the context for the test
    test_emul_save_current_test (frame->addr);

    // We handled this error
    return 0;
  }

  // This is not an error packet that is the intended result of the test; we
  // don't handle it.

  return -1;
}

/**
 * Initialise the communication test running on the board with address addr,
 * using protocol proto.
 *
 * To be called when the "start comm. test" frame is received.
 *
 * Initialises states and attaches handlers to all types of frames and errors we
 * expect to see.
 */
void init_comm_test (const uint8_t addr, const uint8_t proto) {
  struct s_frame dummy;

  // Register the proto handler for the test
  reg_handler (addr, proto, parse_comm_test);
  // Register the error handler for the test
  reg_err_handler (addr, proto, parse_comm_test_err);
  if (addr != 0) {
    // Register the handler that parses management protocol frames from
    // the board to which this board is connected
    reg_handler (addr - 1, 0, parse_comm_test_manag);
  }
  
  // We simply discard the management frame (it triggered all this processing in
  // the first place)
  set_comm_emul_buf (&dummy);
  // Set the context for the test
  test_emul_set_current_test (addr);
  // Emulate the start of the test
  start_comm_test (0);
  // Save the context for the test
  test_emul_save_current_test (addr);
}

/**
 * End the communication test and inspect result.
 *
 * Unregisters all handlers registered by init_comm_test() and destroys the test
 * environment associated with this test.
 *
 * @returns 0 when everything is okay, -1 when an incorrect test frame was
 * received earlier, and -2 when we were still expecting more frames.
 *
 * Precondition: the test is correctly initialized.
 */
int end_comm_test (const uint8_t addr) {
  struct comm_test_var *test_var;
  struct test_data_item *current_test;
  int retval;

  // Get data for this test
  current_test = test_emul_get_data (addr);

  // Calling this function for a non-running test is an error
  assert (current_test != NULL);

  test_var = (struct comm_test_var*) ((void*) current_test->test_data);

  if (test_var->len == MAX_FRAME_SIZE + 1 && test_var->bitpos == 7 && current_test->status == 0) {
    // Succesful: ended at the right moment, and data was okay
    retval = 0;

  } else if (current_test->status == 0) {
    // Test ended prematurely
    retval = -2;

  } else {
    // Data was incorrect at some point
    retval = -1;
  }

  // The test is identified only by the address (always at most 1 test
  // running), so we can delete it without knowing the proto number.

  // Unregister proto handler for the test
  unreg_handler (addr, PROTO_ANY, parse_comm_test);
  // Unregister the error handler for the test
  unreg_err_handler (addr, PROTO_ANY, parse_comm_test_err);

  if (addr != 0) {
    // Unregister the handler that parses management protocol frames from
    // the board to which this board is connected
    unreg_handler (addr - 1, 0, parse_comm_test_manag);
  }

  return retval;
}
